/*
    WOCFS - Write-Only enCrypted File System
    Copyright (C) 2011, Michael Rao

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
     any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "utils.hpp"
#include <regex.h>
#include <signal.h>
#include <pthread.h>
#include "opttree.hpp"

#include <zlib.h>

#ifdef HAVE_SETXATTR
#include <sys/xattr.h>
#endif

#define MAXBUFF (1024*64)

/*
  TODO:
  - option compression

  - compression :
  -- si grand buffer, tester un petit bout d'abord 
  -- en dessous de 100-X %, passer de 2 en 1 au bout de Y ecritures ou Z bytes 
  -- repasser 1 en 2 apres un certain temps. 
  --> compression qui echoue doit pas depasser x % de l'ensemble des bits ecrits
  -- opt dico fixe

  - fichier hint pour major/minor
  - logs

  options:
  - option atonce 
  - option preferer couper a '\n' ?
*/

#define FUSE_USE_VERSION 26
//#define _FILE_OFFSET_BITS 64

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if 0
#ifdef linux
/* For pread()/pwrite() */
#define _XOPEN_SOURCE 500
#endif
#endif

#include <fuse.h>

int vlog=2;

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t saver_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t saver_cond = PTHREAD_COND_INITIALIZER;

void th_lock() {
  //printf("th_lock...\n");
  pthread_mutex_lock(&mutex);
  //printf("th_lock... OK\n");
}

void th_unlock() {
  //printf("th_unlock...\n");
  pthread_mutex_unlock(&mutex);
  //printf("th_unlock... OK\n");
}

pthread_t th_th_s;

gcry_sexp_t pub_key;

opt_t default_opts;
mopt_t default_mopts;

#define WO_ZLIB (1LL<<32)
#define WO_PLAIN (2LL<<32)

ll_t global_opts=(3LL<<32);

struct hash_t {
  unsigned char t[AESKEYSIZE];
  ll_t i;

  void next() {
    unsigned char tp[AESKEYSIZE];
    gcry_md_hash_buffer(GCRY_MD_SHA256,tp,t,AESKEYSIZE);
    memcpy(t,tp,AESKEYSIZE);
    i++;
  }
  
  void rand() {
    gcry_randomize(t,AESKEYSIZE,GCRY_STRONG_RANDOM);
    i=0;
  }
};

static int incrsize=1024;

struct file_t 
{
  void clear() {
    bcap=bsize=0;
    roff=0;
    bf=NULL;
    path=NULL;
    writen=0;
    maj=mino=0;
    compress=2;
  }
  file_t() {clear();}
  file_t(const char *p) {clear();path=strdup(p);}
  void forget();
  ~file_t() {
    assert(bsize==0);
    if(path) free(path);
    if(bf) free(bf);
  }
  void rename(const char* p) {
    if(path) free(path);
    path=strdup(p);
  }
  void init(const char* p);

  file_t(const file_t &o) {
    memcpy(this,&o,sizeof(file_t));
    if(o.path) path=strdup(o.path);
      if(o.bf) {
	bf=(char*)malloc(o.bcap);
	memcpy(bf,o.bf,bsize);
      }
  }
  file_t &operator=(const file_t &o) {
    if(this!=&o) {
      memcpy(this,&o,sizeof(file_t));
      if(o.path) path=strdup(o.path);
      if(o.bf) {
	bf=(char*)malloc(o.bcap);
	memcpy(bf,o.bf,bsize);
      }
    }
    return *this;
  }

  int writeondisk(char *bf,ll_t bsize);
  bool flush(); // write on disk if it has to bee writen
  bool finish(); // write all buffer on disk
    
  void release() { // when the file is closed
    ll_t opts=getopt(path);
    if(opts&A_FINISHONRELEASE) {
      finish();
    } else {
      flush();
    }
  }

  void append(const char *buff,ll_t size);
  bool pwrite(ll_t pos,const char *buff,ll_t size) {
    if(vlog>3) printf("pwrite pos=%Ld size=%Ld roff=%Ld bsize=%d\n",pos,size,roff,bsize);
    if(pos<roff+writen)
      return 0;
    pos-=roff+writen;
    if(pos<bsize) 
      return 0;
    append(buff,size);
    flush();
    return 1;
  }
  
  char* path;
  //int open;
  char *bf;
  int bcap; // taille de bf
  int bsize; //taille ecrit dans bf
  ll_t roff; // relative offset of bf
  
  ll_t time;

  ll_t timeout;
  ll_t maxbuff;
  //ll_t opts;
  
  ll_t writen;

  ll_t maj; // de l'ancien block
  ll_t mino; // de l'ancien block

  int compress; 
  //compress ratio..
};

struct wocfs_opt_t
{
  char *basedir;
};

wocfs_opt_t wocfs_opts;

char *basedir;
char *dir;
char *ddir;
char *pub_key_fn;
char *mainfile_fn;
char *checkfile_fn;

hash_t oth;
ll_t maj;
int maxfiles;

/*
#define A_ORDWR 64 //OK if open with O_RDWR, but fails if read()
#define A_FINISHONRELEASE 128
#define A_FAKEREAD 256
#define A_P_CREATE 16384
*/

void setdefaultopts() {
  default_opts.opt=A_UNLINK+A_RENAME+A_CREATE+A_MKDIR+A_RMDIR+A_TRUNC+A_P_UNLINK+A_P_RENAME+A_P_TRUNC+A_P_SYMLNK;
  default_opts.timeout=5;
  default_opts.maxbuffer=4096;

  maxfiles=-1;
  basedir=NULL;
  maj=0;
  dir=NULL;
  pub_key_fn=strdup("wocfs.pub_key");
  mainfile_fn=strdup("wocfs.main");
  checkfile_fn=strdup("wocfs.check");
}
  
map<string,file_t> files;
multimap<ll_t,file_t *> timeouts;
map<file_t*,multimap<ll_t,file_t*>::iterator > rtimeouts;

 void to_remove(file_t *i) 
 {
   map<file_t*,multimap<ll_t,file_t* >::iterator >::iterator it2=rtimeouts.find(i);
   if(it2==rtimeouts.end()) return;
   timeouts.erase(it2->second);
   rtimeouts.erase(it2);
 }

 void to_insert(file_t* i,ll_t to) 
 {
   if(rtimeouts.find(i)!=rtimeouts.end()) to_remove(i);
   assert(rtimeouts.find(i)==rtimeouts.end());
   multimap<ll_t,file_t*>::iterator it2=timeouts.insert(make_pair(to,i));
   rtimeouts[i]=it2;
 }

 void to_removeall()
 {
   timeouts.clear();
   rtimeouts.clear();
 }


 bool createentry(const char *path) {
   map<string,file_t>::iterator it=files.find(path);
   if(it==files.end()) {
     files[path].init(path);
     return 1;
   }
   return 0;
 }

 void preinit() {
   basedir=NULL;
   if(wocfs_opts.basedir)
     basedir=strdup(wocfs_opts.basedir);
   if(dir==NULL) {
     string d=string(basedir)+string("/encrypted");
     dir=strdup(d.c_str());
   }
   if(ddir==NULL) {
     string d=string(basedir)+string("/plain");
     ddir=strdup(d.c_str());
   }
 }

 void writeinmain(int t) {
   int r;
   int d=0;
   int maxsize=AESKEYSIZE+8+8+8+8+8+8;
   unsigned char bfp[maxsize];

   memcpy(bfp,M_HEAD,4);
   if(t==1)
     writeint(bfp+4,1+M_DATE+M_MAJOR+M_HASH,4);
   else
     writeint(bfp+4,M_CONF+M_DATE+M_MAJOR+M_HASH,4);
   d=8;

   if(t==0) {
     bfp[d++]=M_VER_0;
     bfp[d++]=M_VER_1;
     bfp[d++]=M_VER_2;
     bfp[d++]=M_VER_3;
     writeint(bfp+d,0,4); d+=4;
   }

   ll_t tim=time(NULL)*1000;
   writeint(bfp+d,tim,8);d+=8;
   writeint(bfp+d,maj,8);d+=8;
   //printf("maj=%Ld\n",maj);
   memcpy(bfp+d,oth.t,AESKEYSIZE);d+=AESKEYSIZE;

   assert(d<=maxsize);

   gcry_sexp_t sexp;
   r=gcry_sexp_build(&sexp,NULL,"(data (flags pkcs1) (value %b))",d,bfp);
   if(r) {
     fprintf(stderr,"Error in gcry_sexp_build\n");
     exit(1);
   }
   memset(bfp,0,maxsize);

   gcry_sexp_t csexp;
   r=gcry_pk_encrypt(&csexp, sexp, pub_key);
   if(r) {
     fprintf(stderr,"Error in gcry_pk_encrypt\n");
     exit(1);
   }
   gcry_sexp_release(sexp);

   char bf[100000];
   r=gcry_sexp_sprint(csexp,GCRYSEXP_FMT_ADVANCED,bf,100000);
   assert(r>0 && r<100000);

   //printf("%s",bf);

   size_t len;
   const unsigned char *data;
   gcry_sexp_t c1=gcry_sexp_nth(csexp,1);
   gcry_sexp_t c2=gcry_sexp_nth(c1,1);
   r=gcry_sexp_sprint(c1,GCRYSEXP_FMT_ADVANCED,bf,100000);
   assert(r>0 && r<100000);
   data =(unsigned char *) gcry_sexp_nth_data (c2, 1, &len);

   /*
     printf("%s",bf);
     for(int i=0;i<len;i++) {
     printf("%X",data[i]);
     }
     printf("\n");
   */

   string fn=string(basedir)+string("/")+string(mainfile_fn);
   FILE *out=fopen(fn.c_str(),"a");
   int l2=len;
   fwrite(&l2,1,4,out);
   fwrite(data,1,len,out);
   fclose(out);

   gcry_sexp_release(c2);
   gcry_sexp_release(c1);
   gcry_sexp_release(csexp);

   {
     string fn=string(basedir)+string("/")+string(checkfile_fn);
     FILE *in=fopen(fn.c_str(),"r");
     if(in==0) {
       fprintf(stderr,"creating a new checkfile...\n");
       FILE *out=fopen(fn.c_str(),"w");
       if(out==0) {
	 fprintf(stderr,"FATAL: cannot create %s\n",fn.c_str());
	 exit(1);
       }
       int re=fwrite("0123456789ABCDEF0123456789ABCDEF",1,BLOCKSIZE,out);
       if(re!=BLOCKSIZE) {
	 fprintf(stderr,"FATAL: cannot write in %s\n",fn.c_str());
	 exit(1);
       }
       fclose(out);
       in=fopen(fn.c_str(),"r");
     }
     assert(in);
     char bf[BLOCKSIZE];
     char bf2[BLOCKSIZE];
     int re=fread(bf,1,BLOCKSIZE,in);
     if(re!=BLOCKSIZE) {
       fprintf(stderr,"FATAL: bad checkfile %s\n",fn.c_str());
       exit(1);
     }

     gcry_error_t r;
     gcry_cipher_hd_t hd;
     r=gcry_cipher_open (&hd, GCRY_CIPHER_AES256, GCRY_CIPHER_MODE_ECB,0 & GCRY_CIPHER_SECURE);
     assert(r==0);
     r=gcry_cipher_setkey(hd,oth.t,AESKEYSIZE);
     assert(r==0);
     r=gcry_cipher_encrypt(hd, bf2,BLOCKSIZE,bf,BLOCKSIZE);
     assert(r==0);
     gcry_cipher_close(hd);

     /*
      */

     FILE *out=fopen(fn.c_str(),"w");
     re=fwrite(bf2,1,BLOCKSIZE,out);
     if(re!=BLOCKSIZE) {
       fprintf(stderr,"FATAL: cannot write in %s\n",fn.c_str());
       exit(1);
     }
     fclose(out);
   }

   oth.next();
 }


 void init() {
   oth.rand();

   string fn=string(basedir)+string("/")+string(mainfile_fn);
   FILE *out=fopen(fn.c_str(),"a");
   assert(out);
   maj=ftello(out);
   fclose(out);
   //printf("maj=%Ld\n",maj);
   writeinmain(0);
 }

 void destroy() {
   map<string,file_t>::iterator it;
   for(it=files.begin();it!=files.end();++it) {
     if(vlog>2) printf("finish %s\n",it->first.c_str());
     it->second.finish();
   }
   writeinmain(1);
 }

 string getencfilename(const char *fn)
 {
   return string(dir)+string(fn);
 }

 string getplainfilename(const char *fn)
 {
   return string(ddir)+string(fn);
 }

 #define I_ENC 0
 #define I_PLAIN 1
 #define I_BAD 2
 #define I_BOTH 3

 bool isplain(const char *fn)
 {
   if(0==(global_opts&WO_PLAIN)) return 0;
   string re=getplainfilename(fn);
   struct stat st;
   int res = lstat(re.c_str(), &st);
   if (res >= 0) {
     return true;
   }
   return false;
 }

 bool isenc(const char *fn)
 {
   string re=getencfilename(fn);
   struct stat st;
   int res = lstat(re.c_str(), &st);
   if (res >= 0) {
     return true;
   }
   return false;
 }

 int isplainenc(const char *fn)
 {
   bool e=isenc(fn);
   bool p=isplain(fn);
   if(e==0 && p==0) return I_BAD;
   if(e==0 && p==1) return I_PLAIN;
   if(e==1 && p==0) return I_ENC;
   //fprintf(stderr,"%s is plain and enc !\n",fn);
   if(e==1 && p==1) return I_BOTH;
   assert(0);
   return 0;
 } 

 string getrealfilename(const char *fn)
 {
   bool e=isenc(fn);
   if(isplain(fn)) {
     if(e) {
       fprintf(stderr,"%s is plain and enc !\n",fn);
       return "";
     }
     return getplainfilename(fn);
   }
   if(e)
     return getencfilename(fn);
   return "";
 }

 void file_t::init(const char* p) {
   //TODO? read maj & min
   maj=mino=0;
   timeout=default_opts.timeout;
   maxbuff=default_opts.maxbuffer;
   if(maxbuff<0) maxbuff=MAXBUFF; 
   //opts=getopt(p);
   string re=getencfilename(p);
   struct stat st_data;
   int res = lstat(re.c_str(),&st_data);
   if(res == -1)  {
   } else {
     if(st_data.st_size>0) maj=1;
   }
   rename(p);
 }

 void file_t::forget() {
   to_remove(this);
   bsize=0;
   if(bf) free(bf);
   bf=NULL;
 }


 bool file_t::flush() { 
   //TODO time
   if(bf==NULL) return false;
   if(bsize<maxbuff) return false;
   writeondisk(bf,bsize);
   writen+=bsize;
   free(bf);
   bsize=0;bf=NULL;
   return true;
 }

 bool file_t::finish() { // write all buffer on disk
   to_remove(this);
   if(bf) {
     writeondisk(bf,bsize);
     free(bf);
     writen+=bsize;
     bsize=0;bf=NULL;
   }
   return true;
 }

 int file_t::writeondisk(char *bf_,ll_t bsize_) { 
   char *bf_c=NULL;
   ll_t bsize=bsize_;
   char *bf=bf_;
   bool comp=false;

   /* try compress */
   if((global_opts & WO_ZLIB) && compress>1) {
     z_stream strm;
     int r;
     strm.zalloc=Z_NULL;
     strm.zfree=Z_NULL;
     strm.opaque=Z_NULL;
     if(Z_OK!=deflateInit(&strm, Z_DEFAULT_COMPRESSION)) {
       if(vlog>0) fprintf(stderr,"Error in deflateInit\n");
       goto fail_comp;
     }
     bf_c=(char*)malloc(bsize_);
     strm.next_in=(Bytef*)bf_;
     strm.avail_in=bsize;
     strm.next_out=(Bytef*)bf_c;
     strm.avail_out=bsize;
     r=deflate(&strm,Z_FINISH);
     if(r==Z_STREAM_ERROR) {
       if(vlog>0) fprintf(stderr,"Error deflate returns STREAM_ERROR\n");
       goto fail_comp;
     }
     if(strm.avail_out>0) {
       bf=bf_c;
       bsize=bsize_-strm.avail_out;
     }
     deflateEnd(&strm);
     comp=1;
   fail_comp:
     if(bf!=bf_c) {
       free(bf_c);
       bf_c=NULL;
       comp=0;
     }
   }

   if(vlog>3) printf(" writeondisk buf of size %Ld\n",bsize);
   int sizemax=8*16+bsize+BLOCKSIZE*2;
   unsigned char *b=(unsigned char *)malloc(sizemax);
   unsigned char *bc=(unsigned char *)malloc(sizemax);
   memset(b,0,sizemax);
   
   int tt=0;
   int xxx=4;
   writeint(bc,tt,4);
   xxx+=writecint(bc+xxx,::maj);
   xxx+=writecint(bc+xxx,oth.i);
   
   ll_t flags=BF_PREV_MIN;
   if(maj!=::maj) flags |= BF_PREV_MAJ;
   if(comp) flags |=  BF_COMP;
   int yyy=4;
   writeint(b,flags,4);
   if(flags& BF_PREV_MAJ)
     yyy+=writecint(b+yyy,maj);
   if(flags& BF_PREV_MIN)
     yyy+=writecint(b+yyy,mino);
   //dates...
   yyy+=writecint(b+yyy,bsize);
   //SALT?
   memcpy(b+yyy,bf,bsize);yyy+=bsize;
   yyy+=2; // at least two 0, for check
   
   gcry_error_t r;
   gcry_cipher_hd_t hd;
   r=gcry_cipher_open (&hd, GCRY_CIPHER_AES256, GCRY_CIPHER_MODE_CBC,0 & GCRY_CIPHER_SECURE);
   assert(r==0);
   
   /*
     for(int i=0;i<AESKEYSIZE;i++) {
     printf("%02X",int(oth.t[i]));
     }
     printf(" << KEY\n");
   */
   
   r=gcry_cipher_setkey(hd,oth.t,AESKEYSIZE);
   assert(r==0);
   unsigned int nb=(yyy+BLOCKSIZE-1)/BLOCKSIZE; //nbr of blocks
   
   r=gcry_cipher_encrypt(hd,bc+xxx,BLOCKSIZE*nb,b,BLOCKSIZE*nb);
   assert(r==0);
   gcry_cipher_close(hd);
   
   tt=nb*BLOCKSIZE+xxx-4;
   writeint(bc,tt,4);
  
  string re=getencfilename(path);
  int fd = open(re.c_str(), O_APPEND|O_CREAT|O_WRONLY, 0600);

  maj=::maj;
  mino=oth.i;

  int r2=write(fd,bc,tt+4);
  assert(r2==tt+4);
  close(fd);
  oth.next();
  if(bf_c) {
    free(bf_c);
  }
  free(b);
  free(bc);
  return 0;
}

  
void file_t::append(const char *buff,ll_t size) {
  if(vlog>3) printf("append buf of size %Ld\n",size);
  if(bf==NULL) {
    this->time=::time(NULL);
    if(timeout>0) {
      to_insert(this,this->time+timeout);
      
      //wake up saver thread:
      //printf("wake up saver thread\n");
      pthread_mutex_lock(&saver_mutex);
      //printf("- locked\n");
      pthread_cond_signal(&saver_cond);
      //printf("- signaled\n");
      pthread_mutex_unlock(&saver_mutex);
      //printf("- unlocked\n");
    }
    bcap=size+incrsize;
    bf=(char*)malloc(bcap);
    bsize=0;
  }
  if(size>bcap-bsize) {
    bcap=size+bsize+incrsize;
    bf=(char*)realloc(bf,bcap);
  }
  memcpy(bf+bsize,buff,size);
  bsize+=size;
}


void *th_saver(void*);

static void *wocfs_init(struct fuse_conn_info *conn) 
{
  if(vlog>1) printf("INIT %p\n",conn);
  init();
  //pthread_create(&th_th_s,NULL,th_saver,NULL);
  return NULL;
}

static void wocfs_destroy(void *userdata) 
{
  if(vlog>1) printf("DESTROY %p\n",userdata);
  pthread_cancel(th_th_s);
  //printf("canceled...\n");
  pthread_join(th_th_s,NULL);
  //printf("joined...\n");
  th_lock();
  to_removeall();
  destroy();
  th_unlock();  
  if(vlog>1) printf("DESTROYED\n");
}

static int wocfs_mkdir(const char *path,
		      mode_t mode) 
{
  ll_t opts=getopt(path);
  if((opts&A_MKDIR)==0) return -EPERM;

  string re=getencfilename(path);
  int res = mkdir(re.c_str(), mode); //|S_IRUSR|S_IWUSR|S_IXUSR);
  //MODE ?
  if(res == -1)
    return -errno;

  if(0==(global_opts&WO_PLAIN)) return 0;
  
  re=getplainfilename(path);
  res = mkdir(re.c_str(), mode); //|S_IRUSR|S_IWUSR|S_IXUSR);
  //MODE ?
  if(res == -1)
    return -errno;

  return 0;
}

static int wocfs_getattr(const char *path,struct stat *st_data) 
{
  int i=isplainenc(path);
  if(i==I_ENC || i==I_BOTH) {
    ll_t opt=getopt(path);
    string re=getencfilename(path);
    int res = lstat(re.c_str(), st_data);
    if(res == -1)
      return -errno;
    if(S_ISREG(st_data->st_mode)) {
      if(i==I_BOTH) return -ENOENT;
      th_lock();
      if(0==(opt&A_FAKEREAD))
	st_data->st_mode&=~0555; // no read, no exec
      map<string,file_t>::iterator it=files.find(path);
      if(it!=files.end()) {
	st_data->st_size=it->second.writen+it->second.bsize;
      } else {
	st_data->st_size=0;
      }
      th_unlock();
      return 0;
    }
    return 0;
  }
  if(i&I_BAD) return -ENOENT;
  if(i&I_PLAIN) {
    string re=getplainfilename(path);
    int res = lstat(re.c_str(), st_data);
    if(res == -1)
      return -errno;
    return 0;    
  }
  assert(0);
  return -ENOENT;
}

static int wocfs_access(const char *path, int mode) 
{
  int i=isplainenc(path);
  if(i==I_ENC || i==I_BOTH) {
    struct stat st_data;
    string re=getencfilename(path);
    int res=lstat(re.c_str(), &st_data);
    if(S_ISREG(st_data.st_mode)) {
      if(i==I_BOTH) return -ENOENT;
      // Check if it's write access and deny other operation
      if (mode & ~W_OK) 
	return -EPERM;
    }
    res = access(re.c_str(), mode);
    if (res == -1 )
      return -errno;
    return res;
  }
  if(i&I_BAD) return -ENOENT;
  if(i&I_PLAIN) {
    string re=getplainfilename(path);
    int res = access(re.c_str(), mode);
    if (res == -1 )
      return -errno;
    return res;
  }
  assert(0);
  return -ENOENT;
}

static int wocfs_readlink(const char *path, char *buf, size_t size)
{
  int res;
  
  string re=getplainfilename(path);
  res = readlink(re.c_str(), buf, size - 1);
  if (res == -1)
    return -errno;
  
  buf[res] = '\0';
  return 0;
}

static int wocfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
			 off_t offset, struct fuse_file_info *fi)
{
  DIR *dp,*dp2;
  struct dirent *de;
  
  (void) offset;
  (void) fi;

  //fprintf(stderr,"readdir %s %Ld\n",path,ll_t(offset));
  
  string re=getencfilename(path);
  // fprintf(stderr,"readdir end: %s\n",re.c_str());
  dp = opendir(re.c_str());
  //dp = opendir("/");
  if (dp == NULL)
    return -errno;

#if 1

  if(isplain(path)) {
    string re=getplainfilename(path);
    dp2 = opendir(re.c_str());
    if (dp2 == NULL) {
      closedir(dp);
      return -errno;
    }
    
    while ((de = readdir(dp2)) != NULL) {
      struct stat st;
      memset(&st, 0, sizeof(st));
      st.st_ino = de->d_ino;
      st.st_mode = de->d_type << 12;
      if(de->d_type==DT_REG || de->d_type==DT_LNK) {
	//fprintf(stderr,"P %s %X\n",de->d_name,(int)st.st_mode);
	if (filler(buf, de->d_name, &st, 0))
	  break;
      }
    }
    
    closedir(dp2);
  }

  //return 0;
#endif
  
  while ((de = readdir(dp)) != NULL) {
    struct stat st;
    memset(&st, 0, sizeof(st));
    st.st_ino = de->d_ino;
    st.st_mode = de->d_type << 12;
    //if(0==strcmp(de->d_name,"..")) continue;
    //if(0==strcmp(de->d_name,".")) continue;
    if(de->d_type==DT_REG || de->d_type==DT_DIR) 
      {
	//fprintf(stderr,"E %s %X\n",de->d_name,(int)st.st_mode);
	if (filler(buf, de->d_name, &st, 0))
	  break;
      }
  }
  
  closedir(dp);

  return 0;
}

static int wocfs_mknod(const char *path, mode_t mode, dev_t rdev)
{
  if(mode!=S_IFREG)
    return -EPERM;
  
  string re=getencfilename(path);
  int res = open(re.c_str(), O_CREAT | O_EXCL | O_WRONLY, mode);
  if (res >= 0)
    res = close(res);
  if (res == -1)
    return -errno;
  return 0;
  
  //res = mknod(path, mode, rdev);
}

static int wocfs_rename(const char *from, const char *to) 
{
  int i=isplainenc(from);
  if(i&I_BAD) return -ENOENT;
  if(i&I_PLAIN) {
    ll_t opts=getopt(from);
    if((opts&A_P_RENAME)==0) return -EPERM;

    string rf=getplainfilename(from);
    string rt=getplainfilename(to);
    
    int res = rename(rf.c_str(),rt.c_str());
    if (res == -1)
      return -errno;
    
    return 0;
  }

  ll_t opts=getopt(from);
  if((opts&A_RENAME)==0) return -EPERM;

  //VERIF verif que c'est pas dans files open
  th_lock();
  map<string,file_t>::iterator it=files.find(from);
  if(it!=files.end()) {
    files[to]=it->second;
    files[to].rename(to);
    it->second.forget();
    files.erase(it);
  }
  th_unlock();

  string rf=getencfilename(from);
  string rt=getencfilename(to);

  int res = rename(rf.c_str(),rt.c_str());
  if (res == -1)
    return -errno;
  
  return 0;
  
  //return -EPERM;
}

static int wocfs_chmod(const char *path, mode_t mode) 
{
  int i=isplainenc(path);
  if(i&I_BAD) return -ENOENT;
  if(i&I_PLAIN) {
    string re=getplainfilename(path);
    int res = chmod(re.c_str(), mode);
    if (res == -1)
      return -errno;
    
    return 0;
  }
  string re=getencfilename(path);
  int res = chmod(re.c_str(), mode);
  if (res == -1)
    return -errno;
  
  return 0;
}

static int wherecreate(const char *path)
{
  ll_t opts=getopt(path);
  if((opts&A_CREATE)==0
     && (opts&A_P_CREATE)==0) return I_BAD;
  if((opts&A_CREATE)!=0
     && (opts&A_P_CREATE)!=0) return I_BAD;
  if((opts&A_CREATE)!=0) return I_ENC;
  if((opts&A_P_CREATE)!=0) return I_PLAIN;

  assert(0);
  return -1;
}


static int wocfs_create(const char *path,
		       mode_t mode,
		       struct fuse_file_info *fi) 
{
  int i=wherecreate(path);
  printf("create %d\n",i);
  if(i==I_ENC) {
    // Create the file
    string re=getencfilename(path);
    int fd = open(re.c_str(), fi->flags, mode|S_IWUSR);
    if (fd == -1)
      return -errno;
    
    close(fd);
    //fi->fh = fd;
    return 0;
  }

  // Create the file
  if(i==I_PLAIN) {
    string re=getplainfilename(path);
    int fd = open(re.c_str(), fi->flags, mode);
    if (fd == -1)
      return -errno;
    
    close(fd);
    //fi->fh = fd;
    return 0;
  }
  return -EPERM;
}

static int wocfs_open(const char *path, struct fuse_file_info *fi)
{
  int i=isplainenc(path);
  printf("create (open) %d\n",i);
  if(i==I_BOTH) {
    if(0==(fi->flags&O_CREAT)) 
      return -EPERM;
    return -ENOENT;
  }
  int j=-1;
  if(i==wherecreate(path)) {
    j=wherecreate(path);
  }

  if((i==I_PLAIN) || (i==I_BAD && j==I_PLAIN)) {
    string re=getplainfilename(path);
    int res = open(re.c_str(),fi->flags);
    if (res == -1)
      return -errno;
    close(res);
    return 0;
  }
  if((i==I_ENC) || (i==I_BAD && j==I_ENC)) {
    ll_t opts=getopt(path);
    if(((opts&A_FAKEREAD)==0)  && 
       ((fi->flags&3)==O_RDONLY
	|| ((fi->flags&3)==O_RDWR && (opts&A_ORDWR)==0))) {
      if(vlog>2) printf("open O_RDONLY or O_RDWR without A_ORDWR: return error\n");
      return -EPERM;
    }
    if((opts&A_CREATE)==0 && (fi->flags&O_CREAT)) return -EPERM;
    if((opts&A_TRUNC)==0 && (fi->flags&O_TRUNC)) return -EPERM;
    
    string re=getencfilename(path);
    int res = open(re.c_str(),fi->flags);
    if (res == -1)
      return -errno;
    
    close(res);
    return 0;
  }
  assert(i==I_BAD);
  return -EPERM;
}

static int wocfs_write(const char *path, const char *buf, size_t size, off_t offset, struct fuse_file_info *finfo) 
{
  int i=isplainenc(path);
  if(i&I_BAD) return -ENOENT;
  if(i==I_PLAIN) {
    int fd;
    int res;
    
    (void) finfo;
    string re=getplainfilename(path);
    fd = open(re.c_str(), O_WRONLY);
    if (fd == -1)
      return -errno;
    
    res = pwrite(fd, buf, size, offset);
    if (res == -1)
      res = -errno;
    
    close(fd);
    return res;

    /*
      int res;
      res = pwrite(finfo->fh, buf, size, offset);
      if ((res==-1)||(res != size)) return -errno;
      return res;
    */
  }

  // VERIF
  th_lock();
  createentry(path);
  bool r=files[path].pwrite(offset,buf,size);
  th_unlock();
  if(r) 
    return size;
  else 
    return -EPERM;
}

static int wocfs_release(const char *path, struct fuse_file_info *finfo) 
{
  th_lock();
  map<string,file_t>::iterator it=files.find(path);
  if(it!=files.end()) {
    it->second.release();
  }
  th_unlock();
  return 0;
}

static int wocfs_unlink(const char *path) 
{
  int i=isplainenc(path);
  if(i&I_BAD) return -ENOENT;
  if(i==I_PLAIN) {
    ll_t opts=getopt(path);
    if((opts&A_P_UNLINK)==0) return -EPERM;  

    string re=getplainfilename(path);
    int res = unlink(re.c_str());
    if (res == -1)
      return -errno;
    
    return 0;
    
  }

  ll_t opts=getopt(path);
  if((opts&A_UNLINK)==0) return -EPERM;

  th_lock();
  map<string,file_t>::iterator it=files.find(path);
  if(it!=files.end()) {
    files[path].forget();
    files.erase(it);
  }
  th_unlock();
  
  string re=getencfilename(path);
  int res = unlink(re.c_str());
  if (res == -1)
    return -errno;
  
  return 0;
}

static int wocfs_rmdir(const char *path) 
{
  ll_t opts=getopt(path);
  if((opts&A_RMDIR)==0) return -EPERM;

  string re=getencfilename(path);
  int res = rmdir(re.c_str());
  if (res == -1)
    return -errno;

  re=getplainfilename(path);
  res = rmdir(re.c_str());
  
  return 0;
}

static int wocfs_read(const char *path,
		      char *buf,
		      size_t size,
		      off_t offset,
		      struct fuse_file_info *finfo) 
{
  int i=isplainenc(path);
  if(i&I_BAD) return -ENOENT;
  if(i==I_PLAIN) {
    int fd;
    int res;
    
    (void) finfo;
    string re=getplainfilename(path);
    fd = open(re.c_str(), O_RDONLY);
    if (fd == -1)
      return -errno;
    
    res = pread(fd, buf, size, offset);
    if (res == -1)
      res = -errno;
    
    close(fd);
    return res;
  }

  ll_t opts=getopt(path);
  if(((opts&A_FAKEREAD)==0)) return -EPERM;
  
  memset(buf,'\n',size);

  return size;
}

static int wocfs_chown(const char *path,
		      uid_t uid, gid_t gid) 
{
  string re=getrealfilename(path);
  if(re=="") return -ENOENT;
  int res=lchown(re.c_str(), uid, gid);
  if (res==-1)
    return -errno;
  
  return 0;
}

static int wocfs_truncate(const char *path,
			 off_t size) 
{
  int i=isplainenc(path);
  if(i&I_BAD) return -ENOENT;
  if(i==I_PLAIN) {
    ll_t opts=getopt(path);
    if((opts&A_P_TRUNC)==0) return -EPERM;
    
    string re=getplainfilename(path);
    int res=truncate(re.c_str(), size);
    if (res==-1)
      return -errno;
    
    return 0;
  }

  ll_t opts=getopt(path);
  if((opts&A_TRUNC)==0) return -EPERM;

  string re=getencfilename(path);
  int res=truncate(re.c_str(), size);
  if (res==-1)
    return -errno;

  return 0;
}

static int wocfs_utimens(const char *path, const struct timespec ts[2])
{
  struct timeval tv[2];
  tv[0].tv_sec = ts[0].tv_sec;
  tv[0].tv_usec = ts[0].tv_nsec / 1000;
  tv[1].tv_sec = ts[1].tv_sec;
  tv[1].tv_usec = ts[1].tv_nsec / 1000;
  
  string re=getrealfilename(path);
  int res = utimes(re.c_str(), tv);
  if (res == -1)
    return -errno;
  
  return 0;
}

static int wocfs_symlink(const char *from, const char *to)
{
  ll_t opts=getopt(from);
  if((opts&A_P_SYMLNK)==0) return -EPERM;

  {
    string re=getencfilename(to);
    struct stat st;
    int r=lstat(re.c_str(),&st);
    //if(r==-1) printf("%s \n",strerror(errno));
    if(r!=-1 || errno!=ENOENT)
      return -EEXIST;
  }

  string re=getplainfilename(to);
  //printf("from: %s  to: %s\n",from,to);
  int res = symlink(from,re.c_str());
  if (res == -1)
    return -errno;

  return 0;
}


// reste...

static int wocfs_statfs(const char *path, struct statvfs *st_buf) 
{
  string re=getrealfilename(path);
  int res = statvfs(re.c_str(), st_buf);

  if (res == -1) {
    return -errno;
  }
  return 0;
}

static int wocfs_fsync(const char *path, int crap, struct fuse_file_info *finfo) 
{

  return 0;
}

static int wocfs_link(const char *from, const char *to) 
{
  return -EPERM;
}

/***************/

void *th_saver(void *)
{
  ll_t ti;
  
  while(1) {
    //printf("toto saver\n");
    ti=time(NULL);
    int s;
    th_lock();
    if(timeouts.empty()) s=10;
    else s=timeouts.begin()->first-ti;
    th_unlock();
    if(s<0) s=0;
    if(s) {
      printf("sleep at most %d sec\n",s);
      struct timespec ts;
      ts.tv_sec=ti+s+1;
      ts.tv_nsec=0;

      pthread_mutex_lock(&saver_mutex);
      //printf("+ locked\n");
      pthread_cond_timedwait(&saver_cond,&saver_mutex,&ts);
      //printf("+ waked up r=%d\n",r);
      pthread_mutex_unlock(&saver_mutex);
      //printf("+ unlocked\n");
    }
    pthread_testcancel();
    
    ti=time(NULL);
    th_lock();
    multimap<ll_t,file_t* >::iterator it;
    while(!timeouts.empty()) {
      it=timeouts.begin();
      if(it->first>ti) break;
      file_t* it2=it->second;
      printf("saver: to_remove %s\n",it2->path);
      to_remove(it->second);
      printf("saver: finish %s\n",it2->path);
      it2->finish();
      printf("saver: finish OK %s\n",it2->path);
    }
    th_unlock();
    pthread_testcancel();
  }
  return NULL;
}


/***************/

struct fuse_operations wocfs_oper;

enum {
  KEY_HELP,
  KEY_VERSION,
};

void wocfs_usage()
{
  //TODO
  printf("version %d.%d-%d\nRead README.txt !!!\n",M_VER_1,M_VER_2,M_VER_3);
}

#define POPT_MASK 1
#define POPT_BOOL 1

struct wocfs_parse_opt_t {
  const char* arg;
  int type;
  void *ptr;
  ll_t value;
};

wocfs_parse_opt_t parse_opt[] ={
  {"compress",POPT_MASK,&global_opts,WO_ZLIB},
  {"plain",POPT_MASK,&global_opts,WO_PLAIN},
};

static int wocfs_parse_opt(void *data, const char *arg, int key,
			  struct fuse_args *outargs)
{
  switch (key)
    {
    case FUSE_OPT_KEY_NONOPT:
    case FUSE_OPT_KEY_OPT:
      if(f2o(arg,default_mopts)) return 1;
      for(unsigned int i=0;i<sizeof(parse_opt)/sizeof(parse_opt[0]);i++) {
	if(0==strcmp(arg,parse_opt[i].arg)) {
	  if(parse_opt[i].type==POPT_MASK) 
	    (*(ll_t*)parse_opt[i].ptr)|=parse_opt[i].value;
	  if(parse_opt[i].type==POPT_BOOL) 
	    (*(bool*)parse_opt[i].ptr)=1;
	  return 1;
	}
	if(0==strncmp(arg,"no-",3) && 
	   0==strcmp(arg+3,parse_opt[i].arg)) {
	  if(parse_opt[i].type==POPT_MASK) 
	    (*(ll_t*)parse_opt[i].ptr)&=~parse_opt[i].value;
	  if(parse_opt[i].type==POPT_BOOL) 
	    (*(bool*)parse_opt[i].ptr)=0;
	  return 1;
	}
      }
      fprintf(stderr,"Unkonwn option %s\n",arg);
      return 1;
    case KEY_HELP:
      wocfs_usage();
      exit(1);
    default:
      wocfs_usage();
      exit(1);
    }
  return 1;
}

#define WOCFS_OPT(t, p, v) { t, offsetof(struct wocfs_opt_t, p), v }
#undef FUSE_OPT_END
#define FUSE_OPT_END { NULL,0,0 }

static struct fuse_opt opts[] = 
  {
    WOCFS_OPT("dir=%s",basedir,0),
    FUSE_OPT_KEY("-h",          KEY_HELP),
    FUSE_OPT_KEY("--help",      KEY_HELP),
    FUSE_OPT_END
  };

int main(int argc, char *argv[])
{
  if(argc==1 || 
     (argc==2 && (0==strcmp(argv[1],"-h")*strcmp(argv[1],"--help")))) {
    wocfs_usage();
    exit(1);
  }

  setdefaultopts();

  struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
  
  gcry_control(GCRYCTL_DISABLE_SECMEM);
  const char *version=gcry_check_version(NULL);
  if(vlog>1) printf("Libgcrypt version: %s\n",version);
  //gcry_control( GCRYCTL_ENABLE_QUICK_RANDOM, 0 );

  memset(&wocfs_oper,0,sizeof(wocfs_oper));
  wocfs_oper.init	= wocfs_init;
  wocfs_oper.destroy	= wocfs_destroy;
  wocfs_oper.mknod	= wocfs_mknod;
  wocfs_oper.symlink	= wocfs_symlink;
  wocfs_oper.unlink	= wocfs_unlink;
  wocfs_oper.rmdir	= wocfs_rmdir;
  wocfs_oper.rename	= wocfs_rename;
  wocfs_oper.link	= wocfs_link;
  wocfs_oper.chmod	= wocfs_chmod;
  wocfs_oper.chown	= wocfs_chown;
  wocfs_oper.utimens	= wocfs_utimens;
  wocfs_oper.getattr	= wocfs_getattr;
  wocfs_oper.readlink	= wocfs_readlink;
  wocfs_oper.readdir	= wocfs_readdir;
  wocfs_oper.mkdir	= wocfs_mkdir;
  wocfs_oper.truncate	= wocfs_truncate;
  wocfs_oper.open	= wocfs_open;
  wocfs_oper.read	= wocfs_read;
  wocfs_oper.write	= wocfs_write;
  wocfs_oper.statfs	= wocfs_statfs;
  wocfs_oper.release	= wocfs_release;
  wocfs_oper.fsync	= wocfs_fsync;
  wocfs_oper.access	= wocfs_access;
  wocfs_oper.create     = wocfs_create;

  // Parse options
  //int res=
  fuse_opt_parse(&args, &wocfs_opts, opts, wocfs_parse_opt);
  //TODO verif res et options

  preinit();
  default_opts.apply(default_mopts);
  printf("cmd line flags: ");default_mopts.print();
  printf("\n");

  if(vlog>1) fprintf(stderr,"Main directory: %s\n",basedir);

  struct stat stbuf;
  if(stat(basedir,&stbuf)!=0 || !S_ISDIR(stbuf.st_mode)) {
    if(vlog>0) fprintf(stderr,"Main directory %s does not exist or is not a directory.\n",basedir);
    exit(1);
  }

  string edfn=string(dir);
  if(stat(edfn.c_str(),&stbuf)!=0 || !S_ISDIR(stbuf.st_mode)) {
    if(vlog>0) fprintf(stderr,"Encrypted storage directory %s does not exist or is not a directory.\n",edfn.c_str());
    exit(1);
  }

  string pkfn=string(basedir)+string("/")+string(pub_key_fn);
  if(false==read_pub_key(pkfn.c_str(),&pub_key)) {
    if(vlog>0) fprintf(stderr,"Cannot read public key '%s'.\n",pkfn.c_str());
    exit(1);
  }

  if(false==read_pub_key(pkfn.c_str(),&pub_key)) {
    if(vlog>0) fprintf(stderr,"Cannot read public key '%s'.\n",pkfn.c_str());
    exit(1);
  }

  string optfn=string(basedir)+string("/")+string("wocfs.opts");
  readopts(optfn.c_str());
  compregex(opttree_root);  
    
  pthread_create(&th_th_s,NULL,th_saver,NULL);

  return fuse_main(args.argc, args.argv, &wocfs_oper, NULL);
}
